"""

    routr -- routing engine for WebOb based WSGI applications
    =========================================================

    This module provides data structures along with easy-to-use DSLish API to
    define complex routing rules for WSGI applications based on WebOb.

"""

import re

from pkg_resources import iter_entry_points

try:
    from urllib.parse import urlencode
except ImportError:
    from urllib import urlencode

from webob.exc import HTTPException, HTTPBadRequest
from routr.utils import import_string, cached_property
from routr.urlpattern import URLPattern
from routr.exc import (
    NoMatchFound, NoURLPatternMatched, RouteGuarded,
    MethodNotAllowed, RouteConfigurationError, InvalidRoutePattern,
    RouteReversalError)


__all__ = (
    'Configuration', 'route', 'include', 'plug', 'Trace',
    'Route', 'Endpoint', 'RouteGroup', 'HTTPMethod',
    'GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'OPTIONS', 'TRACE', 'PATCH',
    'NoMatchFound', 'RouteConfigurationError')


class HTTPMethod(str):
    """ HTTP method

    Objects of this type represent HTTP method constants. They are also
    callable -- work as a shortcut for defining route with corresponding
    method -- ``route(GET, ...)`` equivalent to ``GET(...)``.
    """

    def __call__(self, *args, **kwargs):
        return route(self, *args, **kwargs)


GET     = HTTPMethod('GET')
POST    = HTTPMethod('POST')
PUT     = HTTPMethod('PUT')
DELETE  = HTTPMethod('DELETE')
HEAD    = HTTPMethod('HEAD')
OPTIONS = HTTPMethod('OPTIONS')
TRACE   = HTTPMethod('TRACE')
PATCH   = HTTPMethod('PATCH')


class Trace(object):
    """ Result of route matching

    Represents a trace of matched routes towards the last one called endpoint.

    :attr args:
        collected positional arguments
    :attr kwargs:
        collected keyword arguments
    :attr routes:
        a list of matched routes with the endpoint route being the last one
    :attr endpoint:
        matched endpoint route
    """

    def __init__(self, args, kwargs, routes, payload=None):
        self.__dict__['payload'] = payload or {
            'args': args,
            'kwargs': kwargs,
            'routes': routes
        }

    @property
    def endpoint(self):
        return self.routes[-1] if self.routes else None

    @property
    def target(self):
        return self.endpoint.target

    def annotation(self, name, default=None):
        for route in self.routes:
            if name in route.annotations:
                return route.annotations[name]
        return default

    def __add__(self, tr):
        args = self.args + tr.args
        kwargs = dict(self.kwargs)
        kwargs.update(tr.kwargs)
        routes = list(self.routes)
        routes.extend(tr.routes)
        payload = dict(self.payload)
        payload.update(tr.payload)
        payload.update({
            'args': args,
            'kwargs': kwargs,
            'routes': routes})
        return self.__class__(args, kwargs, routes, payload)

    def __setattr__(self, name, value):
        self.payload[name] = value

    def __getattr__(self, name):
        if not name in self.payload:
            raise AttributeError(name)
        return self.payload[name]


class Route(object):
    """ Base class for routes

    :param guards:
        a list of guards
    :param pattern:
        pattern for URL pattern
    :param url_pattern_cls:
        class which should be used for URL pattern matching (default to
        :class:`.urlpattern.URLPattern`
    :param annotations:
        various annotations
    """

    url_pattern_cls = None

    def __init__(self, guards, pattern, url_pattern_cls=None, **annotations):
        self.guards = guards
        self._pattern = pattern
        self.url_pattern_cls = url_pattern_cls
        self.annotations = annotations

    @cached_property
    def pattern(self):
        """ Compiled pattern"""
        return self.compile_pattern(self._pattern)

    def compile_pattern(self, pattern):
        if not pattern:
            return None
        if not pattern.startswith('/'):
            pattern = '/' + pattern
        return (self.url_pattern_cls or URLPattern)(pattern)

    def match_pattern(self, path_info):
        """ Match ``path_info`` against route's ``pattern``"""
        if self.pattern is None:
            if not path_info or path_info == '/':
                return '', ()
            raise NoURLPatternMatched()
        return self.pattern.match(path_info)

    def match_guards(self, request, trace):
        """ Match ``request`` against route's ``guards`` and accumulate result
        in ``trace``
        """
        for guard in self.guards:
            trace = guard(request, trace) or trace
        return trace

    def __call__(self, request):
        """ Try to match route against ``request``

        If no route was matched the :class:`routr.exc.NoMatchFound` exception
        will be raised.

        :param request:
            :class:`webob.Request` object to match route against
        """
        return self.match(request.path_info, request)

    def match(self, path_info, request):
        """ Match ``request`` against route

        return:
            trace object which accumulate ``*args`` and ``**kwargs``
        :rtype:
            :class:`.Trace`

        :raises routr.exc.NoURLPatternMatched:
            if no route was matched by URL
        :raises routr.exc.RouteGuarded:
            if route was guarded by one or more guards
        :raises routr.exc.MethodNotAllowed:
            if method isn't allowed for matched route
        """
        raise NotImplementedError()

    def reverse(self, name, *args, **kwargs):
        """ Reverse route with ``name`` using ``*args`` as pattern parameters
        and ``**kwargs`` as query string parameters

        :raises routr.exc.RouteReversalError:
            if no reversal can be computed for given arguments
        """
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()


class Endpoint(Route):
    """ Endpoint route

    Associated with some object ``target`` which will be returned in case of
    successful match and a ``method`` which matches against request's method.

    Additional to :class:`.Route` params are:

    :param target:
        object to associate with route
    :param method:
        HTTP method associate with route
    :param name:
        optional name, should be provided if reversal of this route is needed,
        otherwise ``None`` is allowed
    """

    def __init__(self, target, method, name, guards, pattern, **annotations):
        super(Endpoint, self).__init__(guards, pattern, **annotations)
        self.target = target
        self.method = method
        self.name = name

    def match_method(self, request):
        if self.method != request.method:
            raise MethodNotAllowed()

    def match(self, path_info, request):
        path_info, args = self.match_pattern(path_info)
        if path_info:
            raise NoURLPatternMatched()
        self.match_method(request)
        trace = Trace(args, {}, [self])
        trace = self.match_guards(request, trace)
        return trace

    def reverse(self, name, *args, **kwargs):
        if name != self.name:
            raise RouteReversalError("no route with name '%s'" % name)
        url = self.pattern.reverse(*args) if self.pattern else '/'
        if kwargs:
            url += '?' + urlencode(kwargs)
        return url

    def __iter__(self):
        return iter([self])

    def __repr__(self):
        return '%s(target=%r, guards=%r, pattern=%r)' % (
            self.__class__.__name__, self.target, self.guards,
            self.pattern.pattern if self.pattern else None)

    __str__ = __repr__


class RouteGroup(Route):
    """ Route which represents a group of other routes

    Can have its own ``guards`` and a URL ``pattern``.

    Additional to :class:`.Route` params are:

    :param routes:
        a list of :class:`Route` objects
    """

    def __init__(self, routes, guards, pattern, url_pattern_cls=None,
                 **annotations):
        super(RouteGroup, self).__init__(
            guards, pattern, url_pattern_cls=url_pattern_cls, **annotations)
        self.routes = routes

    def index(self):
        """ Return mapping from route name to actual route"""
        idx = {}
        for r in self.routes:
            if isinstance(r, Endpoint) and r.name:
                if r.name in idx:
                    raise RouteConfigurationError(
                        "route this name '%s' already defined")
                if self.pattern or r.pattern:
                    idx[r.name] = self.pattern + r.pattern
                else:
                    idx[r.name] = (self.url_pattern_cls or URLPattern)('/')
            elif isinstance(r, RouteGroup):
                ridx = r.index()
                if set(ridx) & set(idx):
                    raise RouteConfigurationError(
                        "route this name '%s' already defined")
                for (n, u) in ridx.items():
                    idx[n] = self.pattern + u
        return idx

    @cached_property
    def _cached_index(self):
        return self.index()

    def reverse(self, name, *args, **kwargs):
        if not name in self._cached_index:
            raise RouteReversalError("no route with name '%s'" % name)
        url = self._cached_index[name].reverse(*args)
        if kwargs:
            url += '?' + urlencode(kwargs)
        return url

    def match_pattern(self, path_info):
        """ Match ``path_info`` against route's ``pattern``"""
        if self.pattern is None:
            return path_info, ()
        return self.pattern.match(path_info)

    def match(self, path_info, request):
        path_info, args = self.match_pattern(path_info)
        guarded = []
        trace = Trace(args, {}, [self])
        trace = self.match_guards(request, trace)
        for subroute in self.routes:
            try:
                subtrace = subroute.match(path_info, request)
            except NoURLPatternMatched:
                continue
            except MethodNotAllowed as e:
                guarded.append(RouteGuarded(e, e.response))
                continue
            except RouteGuarded as e:
                guarded.append(e)
                continue
            except HTTPException as e:
                guarded.append(RouteGuarded(e, e))
                continue
            else:
                return ((trace + subtrace)
                        if subtrace is not None and trace is not None
                        else trace or subtrace)
        if guarded:
            # NOTE we raise only last guard failure
            # cause it's more interesting one
            raise guarded[-1]
        raise NoURLPatternMatched()

    def __iter__(self):
        return iter(self.routes)

    def __repr__(self):
        return '%s(routes=%r, guards=%r, pattern=%r)' % (
            self.__class__.__name__, self.routes, self.guards, self.pattern)

    __str__ = __repr__


def include(spec):
    """ Include routes by ``spec``

    :param spec:
        asset specification which points to :class:`.Route` instance
    """
    r = import_string(spec)
    if not isinstance(r, Route):
        raise RouteConfigurationError(
            "route included by '%s' isn't a route" % spec)
    return r


def plug(name):
    """ Plug routes by ``setuptools`` entry points, identified by ``name``

    :param name:
        entry point name to query routes
    """
    routes = []
    for p in iter_entry_points('routr', name=name):
        r = p.load()
        if not isinstance(r, Route):
            raise RouteConfigurationError(
                "entry point '%s' doesn't point at Route instance" % p)
        routes.append(r)
    return RouteGroup(routes, [])


def route(*args, **kwargs):
    """ Directive for configuring routes in application

    :param args:
        ([method,] [pattern,] target) produces endpoint route
        ([method,] [pattern,] *routes) produces route group
    :param kwargs:
        name and guards are treated as name and guards for routes, other
        keyword args a are passed as annotations
    """

    def consume(pred):
        r = []
        while args:
            if not pred(args[0]):
                break
            r.append(args.pop(0))
        return r, args

    args = list(args)

    if not args:
        raise RouteConfigurationError('empty routes')

    method = args.pop(0) if isinstance(args[0], HTTPMethod) else GET
    name = kwargs.pop('name', None)
    url_pattern_cls = kwargs.pop('url_pattern_cls', None)

    if not args:
        raise RouteConfigurationError('empty routes')

    if len(args) == 1 and not isinstance(args[0], Route):
        target = args[0]
        return Endpoint(target, method, name, [], None,
                        url_pattern_cls=url_pattern_cls, **kwargs)

    elif (
            len(args) == 2
            and isinstance(args[0], str)
            and not isinstance(args[1], Route)):
        pattern, target = args
        return Endpoint(target, method, name, [], pattern,
                        url_pattern_cls=url_pattern_cls, **kwargs)

    else:
        if isinstance(args[0], str):
            pattern = args.pop(0)
        else:
            pattern = None

        guards, args = consume(
            lambda d: not isinstance(d, Route) and hasattr(d, '__call__'))

        routes, args = consume(
            lambda d: isinstance(d, Route))
        if routes:
            if url_pattern_cls:
                for r in routes:
                    if r.url_pattern_cls is None:
                        r.url_pattern_cls = url_pattern_cls
            return RouteGroup(routes, guards, pattern,
                              url_pattern_cls=url_pattern_cls, **kwargs)
        elif len(args) == 1:
            target = args[0]
            return Endpoint(target, method, name, guards,
                            pattern, url_pattern_cls=url_pattern_cls, **kwargs)
        elif not args and guards:
            target = guards.pop()
            return Endpoint(target, method, name, guards,
                            pattern, url_pattern_cls=url_pattern_cls, **kwargs)
        else:
            raise RouteConfigurationError(
                "improper usage of 'route' directive")



"""

    routr.tests -- test suite
    =========================

"""

try:
    from unittest2 import TestCase
except ImportError:
    from unittest import TestCase

import six

from webob import Request, exc

from routr import Route, Endpoint, RouteGroup, URLPattern
from routr import route, RouteConfigurationError
from routr import POST, GET
from routr.utils import (
    ImportStringError, positional_args, inject_args, import_string)
from routr.exc import (
    NoURLPatternMatched, RouteGuarded, MethodNotAllowed, RouteReversalError)


__all__ = ()


class TestRouting(TestCase):

    def assertNoMatch(self, r, url):
        if isinstance(url, Request):
            self.assertRaises(NoURLPatternMatched, r, url)
        else:
            self.assertRaises(NoURLPatternMatched, r, Request.blank(url))


class TestRootEnpoint(TestRouting):

    def test_reverse(self):
        r = route('target', name='news')
        self.assertEqual(r.reverse('news'), '/')
        self.assertRaises(RouteReversalError, r.reverse, 'news2')

    def test_match(self):
        r = route('target')
        req = Request.blank('/')
        tr = r(req)
        self.assertEqual(len(tr.routes), 1)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'target'))

    def test_no_match(self):
        def target():
            return 'hello'
        r = route(target)
        req = Request.blank('/news')
        self.assertRaises(NoURLPatternMatched, r, req)


class TestEndpoint(TestRouting):

    def test_reverse(self):
        r = route('news', 'target', name='news')
        self.assertEqual(r.reverse('news'), '/news')
        self.assertRaises(RouteReversalError, r.reverse, 'news2')

        r = route('news/{id:int}/', 'target', name='news')
        self.assertEqual(r.reverse('news', 42), '/news/42/')
        self.assertRaises(RouteReversalError, r.reverse, 'news2')

        r = route('news/{any(login,attach,app_login,access_token)}',
                  'target', name='news')
        self.assertEqual(r.reverse('news', 'aa'), '/news/aa')
        self.assertRaises(RouteReversalError, r.reverse, 'news2')

    def test_match(self):
        r = route('news', 'target')
        req = Request.blank('/news')
        tr = r(req)
        self.assertEqual(len(tr.routes), 1)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'target'))

    def test_no_match(self):
        self.assertNoMatch(
            route('news', 'target'),
            '/new')
        self.assertNoMatch(
            route('news', 'target'),
            '/newsweek')

    def test_method(self):
        r = route(POST, 'news', 'target')
        req = Request.blank('/news', {'REQUEST_METHOD': 'POST'})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'target'))

        self.assertRaises(
            MethodNotAllowed,
            r, Request.blank('/news', {'REQUEST_METHOD': 'DELETE'}))

    def test_param_pattern_int(self):
        r = route('/news/{id:int}/', 'target')
        req = Request.blank('/news/42/')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((42,), {}, 'target'))

        self.assertNoMatch(r, Request.blank('/news/'))
        self.assertNoMatch(r, Request.blank('/news/a/'))
        self.assertNoMatch(r, Request.blank('/news//'))
        self.assertNoMatch(r, Request.blank('/news/122'))

        r = route('/news/{a:int}/{b:int}/{c:int}/', 'target')
        req = Request.blank('/news/42/41/40/')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((42, 41, 40), {}, 'target'))

    def test_param_pattern_string(self):
        r = route('/news/{id:string}/', 'target')

        req = Request.blank('/news/42/')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((six.u('42'),), {}, 'target'))

        req = Request.blank('/news/abcdef-12/')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((six.u('abcdef-12'),), {}, 'target'))

    def test_param_pattern_path(self):
        r = route('/news/{p:path}', 'target')

        req = Request.blank('/news/42/news')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((six.u('42/news'),), {}, 'target'))

        r = route('/news/{p:path}/comments', 'target')

        req = Request.blank('/news/42/news/comments')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((six.u('42/news'),), {}, 'target'))


class TestRouteGroup(TestRouting):

    def test_reverse(self):
        r = route(
            'api',
            route('news', 'news', name='news'),
            route('comments', 'comments', name='comments'))
        self.assertEqual(r.reverse('news'), '/api/news')
        self.assertEqual(r.reverse('comments'), '/api/comments')
        self.assertRaises(RouteReversalError, r.reverse, 'a')

        r = route(
            'api',
            route('news/{id}/', 'news', name='news'),
            route('comments', 'comments', name='comments'))
        self.assertEqual(r.reverse('news', 'hello'), '/api/news/hello/')

        r = route(
            'api',
            route(GET, 'news', name='get-news'),
            route(POST, 'news', name='create-news'))
        self.assertEqual(r.reverse('get-news'), '/api')
        self.assertEqual(r.reverse('create-news'), '/api')

    def test_custom_url_pattern_cls(self):
        from routr.urlpattern import URLPattern

        class MyURLPattern(URLPattern):
            pass

        r = route(
            'api',
            route(GET, 'news', 'news', name='get-news'),
            route(POST, 'news', 'news', name='create-news'),
            url_pattern_cls=MyURLPattern)

        self.assertTrue(isinstance(r.pattern, MyURLPattern))
        for r in r.routes:
            self.assertTrue(isinstance(r.pattern, MyURLPattern))

    def test_reverse_empty_pattern(self):
        r = route(route('news', name='news'))
        self.assertEqual(r.reverse('news'), '/')

    def test_simple(self):
        r = route(
            route('news', 'news'),
            route('comments', 'comments'))

        req = Request.blank('/news')
        tr = r(req)
        self.assertEqual(len(tr.routes), 2)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'news'))

        req = Request.blank('/comments')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'comments'))

        self.assertNoMatch(r, '/newsweeek')
        self.assertNoMatch(r, '/ne')

    def test_group_inexact_pattern(self):
        r = route(
            'news',
            route('{id:int}',
            route('comments', 'view')))
        req = Request.blank('/news/42/comments')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.endpoint.target),
            ((42,), {}, 'view'))

        r = route('news/{id:int}', route('comments', 'view'))
        req = Request.blank('/news/42/comments')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.endpoint.target),
            ((42,), {}, 'view'))

        r = route('news', route('{id:int}/comments', 'view'))
        req = Request.blank('/news/42/comments')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.endpoint.target),
            ((42,), {}, 'view'))

    def test_complex_match(self):
        def news():
            return 'news'

        def comments():
            return 'comments'

        def api_news():
            return 'api_news'

        def api_comments():
            return 'api_comments'

        r = route(
            route(
                'api',
                route('news', 'api_news'),
                route('comments', 'api_comments')),
            route('news', 'news'),
            route('comments', 'comments'))

        req = Request.blank('/news')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'news'))

        req = Request.blank('/comments')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'comments'))

        req = Request.blank('/api/news')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'api_news'))

        req = Request.blank('/api/comments')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'api_comments'))

    def test_by_method(self):
        r = route(
            'api',
            route(GET, 'news_get'),
            route(POST, 'news_post'))

        req = Request.blank('/api', {'REQUEST_METHOD': 'POST'})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'news_post'))

        req = Request.blank('/api')
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'news_get'))

    def test_method_inner(self):
        r = route(
            route('news', 'news'),
            route(GET, 'comments', 'comments_get'),
            route(POST, 'comments', 'comments_post'))

        req = Request.blank('/news', {'REQUEST_METHOD': 'GET'})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'news'))

        req = Request.blank('/news', {'REQUEST_METHOD': 'POST'})
        self.assertRaises(
            RouteGuarded,
            r, req)

        req = Request.blank('/comments', {'REQUEST_METHOD': 'POST'})
        tr = r(req)
        self.assertEqual(
            (tr.args, tr.kwargs, tr.target),
            ((), {}, 'comments_post'))

        req = Request.blank('/comments', {'REQUEST_METHOD': 'DELETE'})
        self.assertRaises(
            RouteGuarded,
            r, req)

        self.assertNoMatch(r, '/newsweeek')
        self.assertNoMatch(r, '/ne')

    def test_guards(self):
        pass


class TestRouteDirective(TestCase):

    def test_root_endpoint(self):
        r = route('myapp.mytarget')
        self.assertEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, 'myapp.mytarget')

    def test_root_endpoint_func(self):
        def target():
            pass
        r = route(target)
        self.assertEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, target)

    def test_root_endpoint_guards(self):
        r = route(id, 'myapp.mytarget')
        self.assertEqual(r.pattern, None)
        self.assertEqual(r.guards, [id])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, 'myapp.mytarget')

    def test_endpoint(self):
        r = route('news', 'myapp.mytarget')
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, 'myapp.mytarget')

    def test_endpoint_func(self):
        def target():
            pass
        r = route('news', target)
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, target)

    def test_endpoint_guards(self):
        r = route('news', id, 'myapp.mytarget')
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [id])
        self.assertIsInstance(r, Endpoint)
        self.assertEqual(r.target, 'myapp.mytarget')

    def test_route_list_no_pattern(self):
        r = route(
            route('news', 'myapp.api.news'),
            route('comments', 'myapp.api.comments'))
        self.assertEqual(r.guards, [])
        self.assertEqual(r.pattern, None)
        self.assertIsInstance(r, RouteGroup)

    def test_route_list_no_pattern_guards(self):
        r = route(
            id,
            route('news', 'myapp.api.news'),
            route('comments', 'myapp.api.comments'))
        self.assertEqual(r.guards, [id])
        self.assertEqual(r.pattern, None)
        self.assertIsInstance(r, RouteGroup)

    def test_route_list(self):
        r = route(
            'api',
            route('news', 'myapp.api.news'),
            route('comments', 'myapp.api.comments'))
        self.assertEqual(r.guards, [])
        self.assertNotEqual(r.pattern, None)
        self.assertIsInstance(r, RouteGroup)

    def test_route_list_guards(self):
        r = route(
            'api',
            id,
            route('news', 'myapp.api.news'),
            route('comments', 'myapp.api.comments'))
        self.assertNotEqual(r.pattern, None)
        self.assertEqual(r.guards, [id])
        self.assertIsInstance(r, RouteGroup)

    def test_invalid_routes(self):
        self.assertRaises(RouteConfigurationError, route)


class TestURLPattern(TestCase):

    def test_int(self):
        p = URLPattern('/a/{id:int}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/42/b/'), ('', (42,)))

    def test_str(self):
        p = URLPattern('/a/{id}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/42/b/'), ('', ('42',)))

        p = URLPattern('/a/{id:str}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/42/b/'), ('', ('42',)))

        p = URLPattern('/a/{id:string}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/42/b/'), ('', ('42',)))

    def test_str_re(self):
        p = URLPattern('/a/{id:str(re=[0-9]+)}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/42/b/'), ('', ('42',)))
        self.assertRaises(NoURLPatternMatched, p.match, '/a/a/b/')

        p = URLPattern('/a/{id:str(re=[0-9a-f]{6})}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/12efa3/b/'), ('', ('12efa3',)))
        self.assertRaises(NoURLPatternMatched, p.match, '/a/a/b/')

    def test_path(self):
        p = URLPattern('/a/{id:path}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/42/43/b/'), ('', ('42/43',)))

    def test_any(self):
        p = URLPattern('/a/{id:any(aaa, bbb, ccc)}/b/')
        self.assertTrue(not p.is_exact)
        self.assertEqual(p.match('/a/aaa/b/'), ('', ('aaa',)))
        self.assertEqual(p.match('/a/bbb/b/'), ('', ('bbb',)))
        self.assertEqual(p.match('/a/ccc/b/'), ('', ('ccc',)))
        self.assertRaises(NoURLPatternMatched, p.match, '/a')
        self.assertRaises(NoURLPatternMatched, p.match, '/a/abc/b/')


class TestPositionalArgs(TestCase):

    def test_func(self):
        def f(a, b, c):
            pass
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

        def f(a, b, c, d=1):
            pass
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

        def f(a, b, c, **kw):
            pass
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

        def f(*args):
            pass
        self.assertEqual(positional_args(f), [])

        def f(*args, **kw):
            pass
        self.assertEqual(positional_args(f), [])

        def f(a, b, c, *args):
            pass
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

        def f(a, b, c, *args, **kw):
            pass
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

    def test_lambda(self):
        f = lambda a, b, c: None
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])
        f = lambda a, b, c, d=1: None
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])
        f = lambda a, b, c, **kw: None
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])
        f = lambda *args: None
        self.assertEqual(positional_args(f), [])
        f = lambda *args, **kw: None
        self.assertEqual(positional_args(f), [])
        f = lambda a, b, c, *args: None
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])
        f = lambda a, b, c, *args, **kw: None
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

    def test_type(self):
        class f(object):
            def __init__(self, a, b, c):
                pass
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

    def test___call__(self):
        class f(object):
            def __call__(self, a, b, c):
                pass
        f = f()
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])

    def test_method(self):
        class f(object):
            def method(self, a, b, c):
                pass
        f = f().method
        self.assertEqual(positional_args(f), ['a', 'b', 'c'])


class TestInjectArgs(TestCase):

    def test_simple(self):
        def f(user, a, request):
            pass
        self.assertEqual(
            inject_args(f, ['a'], user='user', request='request'),
            ['user', 'a', 'request'])

        def f(a, request):
            pass
        self.assertEqual(
            inject_args(f, ['a'], user='user', request='request'),
            ['a', 'request'])

        def f(user, a):
            pass
        self.assertEqual(
            inject_args(f, ['a'], user='user', request='request'),
            ['user', 'a'])


class TestImportString(TestCase):

    def test_error(self):
        self.assertRaises(ImportStringError,
                          import_string,
                          'does_not_exist.exe')
        self.assertRaises(ImportError, import_string, 'does_not_exist.exe')

    def test_success(self):
        module = import_string('six')
        self.assertEqual(module, six)

        klass = import_string('routr.Route')
        self.assertEqual(klass, Route)

    def test_unicode(self):
        module = import_string(six.u('six'))
        self.assertEqual(module, six)

        klass = import_string(six.u('routr.Route'))
        self.assertEqual(klass, Route)



"""

    routr.app -- basic WSGI app as a shortcut for using routr
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

"""

__all__ = ('App',)

class App(object):

    def __init__(self, ):
        self.config = config
        self.routes = self.make_routes()

    def wcat(self, id):
        if not ":" in id:
            raise HTTPNotFound()
        ctype, uri = id.split(":", 1)
        task = Task(uri, ctype)
        try:
            processor = self.config.processors[task.ctype]
        except KeyError:
            raise HTTPNotFound()
        (new_tasks, finished_task) = processor.process(task)
        store_task(finished_task)
        q = runtime.queue()
        for t in new_tasks:
            q.put(json_dumps(t))
        return processor.render(finished_task.struct)

    def make_routes(self):
        return route(
            GET('/{id:path}', self.wcat),
            )

    def __call__(self, environ, start_response):
        request = Request(environ)
        with runtime.new(self.config):
            try:
                trace = self.routes(request)
                args = inject_args(trace.target, trace.args,
                    request=request)

                response = trace.target(*args, **trace.kwargs)

            except NoMatchFound as e:
                response = e.response
            except HTTPError as e:
                response = e

            if not isinstance(response, Response):
                response = Response(json=response)

            return response(environ, start_response)



import os
import sys

from setuptools import find_packages, setup


is_py26 = sys.version_info[:2] == (2, 6)
version = '0.7.1'


setup(
    name='routr',
    version=version,
    description='URL routing made right',
    long_description=open('README').read() + '\n\n' + open('CHANGES').read(),
    author='Andrey Popp',
    author_email='8mayday@gmail.com',
    url='http://routr.readthedocs.org/',
    license='BSD',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    install_requires=list(filter(None, [
        'WebOb >= 1.2b3',
        'six >= 1.3.0',
        'unittest2 == 0.5.1' if is_py26 else None,
    ])),
    include_package_data=True,
    test_suite='routr.tests',
    zip_safe=False,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: WSGI'
    ],
    keywords='routr routes routing webob')
